-- Copyright  2019
-- All Rights Reserved.
-- 
-- Permission to, copy, modify, and distribute this software in any way other than for personal use
-- is not granted without expressed written agreement. 
-- This copyright applies to the all of the code, data and artwork related to the add-on.
--
-- For more information on copyright, Google it.
--
-- Yes, Copyright is a real thing even for World of Warcraft addons.

------------------------------------------------------------------------------------------------------------------
--						  Grid Control							--
------------------------------------------------------------------------------------------------------------------
if AzGridInitialise then return end

local find, len, sub = string.find, string.len, string.sub
local floor = math.floor

local All, Horizontal, Vertical, EndPoint, Axial = 0, 1, 3, 1, 2
local MinLines, MaxLines, MinSpacing, MaxSpacing = 2, 150, 5, 500
local Points = { "LEFT", "RIGHT", "TOP", "BOTTOM" }

local TTStrata = GameTooltip:GetFrameStrata()
local TTLevel = GameTooltip:GetFrameLevel()

local function DisplayTooltip(self)
	local name = self:GetName()
	name = sub(name, find(name, ":")+1, len(name)) 
	if self.Tooltip[name] then
		GameTooltip:SetFrameStrata("TOOLTIP")
		GameTooltip:SetFrameLevel(8)
		GameTooltip:SetOwner(self, "ANCHOR_LEFT ")
		GameTooltip:SetText(self.Tooltip[name])			
	end
end

local function HideTooltip(self)
	GameTooltip:Hide()
	GameTooltip:SetFrameStrata(TTStrata)
	GameTooltip:SetFrameLevel(TTLevel)
end

local function AddTooltips(self, tooltips)
	if not tooltips then return end
	self.Tooltip = tooltips
	self:SetScript("OnEnter", DisplayTooltip)
	self:SetScript("OnLeave", HideTooltip)
end

function AzGridInitialise(f, gridparent, settings, tooltips, point, block)
	local gSettings = settings
	local gTooltips = tooltips
	local OptionsParent = f
	local GridSpacing = gSettings.initspacing

	local DisplayTicks, GridParent

	local function GridSetColor(self)
		if not self then return 1 end -- Min starting line # 1 for processing
		if self.Id > 1 then
			for i=1, 2 do
				self["EndPoint"..i]:GetHighlightTexture():SetColorTexture(unpack(gSettings.colors.EndPoints))
				self["EndPoint"..i].Text:SetTextColor(unpack(gSettings.colors.Texts))
			end
			if self.EndPoint1:GetChecked() then
				self:SetColorTexture(unpack(gSettings.colors.Highlight))
			else
				self:SetColorTexture(unpack(gSettings.colors.Lines))
			end
		else
			self:SetColorTexture(unpack(gSettings.colors.Centers))
		end
	end
	f.GridSetColor = GridSetColor
	
	local function GridClearHightlight(self)
		if not self then return 2 end -- Min starting line # 2 for processing
		if not self.EndPoint1:GetChecked() then return end
		self.EndPoint1:SetChecked(false)
		self.EndPoint2:SetChecked(false)
		GridSetColor(self)
		if DisplayTicks then
			self:Hide()
		end
	end

	local function GridSetVisible(self, visible)
		if not self then return 2 end -- Min starting line # 2 for processing
		if self.Id > 1 and self.Id > GridParent.Count[self.Direction] then
			self:Hide()
			self.EndPoint1:Hide()
			self.EndPoint2:Hide()
			return
		end
		self:SetShown(visible)
		self.EndPoint1:SetShown(visible)
		self.EndPoint2:SetShown(visible)
		if visible and DisplayTicks then
			if self.Id == 1 or not self.EndPoint1:GetChecked() then
				self:Hide()
			end
		end
	end

	local function GridAnchorEndPoints(self)
		self.EndPoint1:ClearAllPoints()
		self.EndPoint2:ClearAllPoints()
		self.EndPoint1:SetPoint(self.EndPoint1.Point, self, self.EndPoint1.Point, 0, 0)
		self.EndPoint2:SetPoint(self.EndPoint2.Point, self, self.EndPoint2.Point, 0, 0)
	end

	local function GridAnchorLine(self, direction, x, y)
		if self.Id then
			local multi = floor(self.Id/2)
			x = x * multi
			y = y * multi
			if self.Id % 2 ~= 0 then
				x = -x
				y = -y
			end
		end
		local C1, C2 = self.EndPoint1, self.EndPoint2
		GridSetColor(self)
		C1.Point = Points[direction] 
		C2.Point = Points[direction+1]
		C1:ClearAllPoints()
		C2:ClearAllPoints()
		GridAnchorEndPoints(self)
		self:SetStartPoint(self.EndPoint1.Point, GridParent, x, y)
		self:SetEndPoint(self.EndPoint2.Point, GridParent, x, y)
		self.EndPoint1.Offset = 0
		if direction == Vertical then
			self.EndPoint2.Offset = -self.EndPoint1:GetTop()
		else
			self.EndPoint2.Offset = self.EndPoint2:GetRight()
		end
		GridSetVisible(self, true)
	end

	local function RunProcess(table, func, value, start)
		local min = func()
		if not start or start < min then
			start = min
		end
		for i=start, #table do
			func(table[i], value)
		end
	end

	local function GridProcessLines(direction, func, value, start, starty)
		if not GridParent then return end
		if direction == All then
			RunProcess(GridParent.Lines[Horizontal], func, value, start)
			RunProcess(GridParent.Lines[Vertical], func, value, starty or start)
		else
			RunProcess(GridParent.Lines[direction], func, value, start)
		end
	end
	f.GridProcessLines = GridProcessLines

	local function GridOnUpdate(self)
		local p, r, t, x, y = self:GetPoint(1)
		if self.x == x and self.y == y then return end
		self.x = x
		self.y = y
		self:ClearAllPoints()
		self.Diametric:ClearAllPoints()
		if self.Direction == Vertical then
			y = 0
			if self == self.Line.EndPoint1 then
				self:SetPoint(p, r, t, x, self.Offset)
				self.Diametric:SetPoint(p, r, t, x, self.Diametric.Offset+self:GetHeight())
				self.Line:SetStartPoint(p, GridParent, x, y)
				self.Line:SetEndPoint(p, GridParent, x, self.Diametric.Offset)
			else
				self:SetPoint(p, r, t, x, self.Offset+self:GetHeight())
				self.Diametric:SetPoint(p, r, t, x, self.Diametric.Offset)
				self.Line:SetStartPoint(p, GridParent, x, self.Diametric.Offset)
				self.Line:SetEndPoint(p, GridParent, x, self.Offset)
			end
		else
			x = 0
			if self == self.Line.EndPoint1 then
				self:SetPoint(p, r, t, self.Offset, y)
				self.Diametric:SetPoint(p, r, t, self.Diametric.Offset-self:GetWidth(), y)
				self.Line:SetStartPoint(p, GridParent, x, y)
				self.Line:SetEndPoint(p, GridParent, self.Diametric.Offset, y)
			else
				self:SetPoint(p, r, t, self.Offset-self:GetWidth(), y)
				self.Diametric:SetPoint(p, r, t, self.Diametric.Offset, y)
				self.Line:SetStartPoint(p, GridParent, self.Diametric.Offset, y)
				self.Line:SetEndPoint(p, GridParent, self.Offset, y)
			end
		end
	end

	local function GridStarMoving(self)
		self.x = -1
		self.y = -1
		OptionsParent.Reset:SetEnabled(true)
		self:SetScript("OnUpdate", GridOnUpdate)
		self:StartMoving()
	end

	local function GridStopMoving(self)
		self:SetScript("OnUpdate", nil)
		self:StopMovingOrSizing()
		GridAnchorEndPoints(self.Line)
	end
	
	local function GridEndpointOnClick(self, button, down)
		if IsShiftKeyDown() then
			self:SetChecked(true)
			GridProcessLines(self.Direction, GridClearHightlight)
			return
		end
		self.Diametric:SetChecked(self:GetChecked())
		GridSetColor(self.Line)
		GridProcessLines(All, GridSetVisible, true)
	end

	local function GridCreateEndPoints(self, direction, id)
		local pad = 0
		local Level = GridParent:GetFrameLevel()
		for i=1, 2 do
			local f
			if id == 1 then
				f = CreateFrame("Frame", nil, GridParent) -- no highlighting/dragging of centre lines
			else
				f = CreateFrame("CheckButton", nil, GridParent)
				f:SetHighlightTexture("Interface\\TargetingFrame\\UI-StatusBar")
				f:SetScript("OnClick", GridEndpointOnClick)
				f.Direction = direction
				f:RegisterForDrag("LeftButton")
				f:SetMovable(true)
				f:SetScript("OnDragStart", GridStarMoving)
				f:SetScript("OnDragStop", GridStopMoving)
			end
			f:SetFrameStrata(GridParent:GetFrameStrata())
			f:SetFrameLevel(Level+1)
			self["EndPoint"..i] = f
			f.Line = self
			f.Text = f:CreateFontString()
			f.Text:SetDrawLayer("ARTWORK", 2)
			f.Text:SetFont("Fonts\\FRIZQT__.TTF", 9)
			local text = tostring(id)	
			if direction == Vertical then
				f:SetSize(12, 15)
				text = gsub(text, "(.)", function(x) return x.."\n" end)
				text = sub(text, 1, len(text)-1)
			else
				f:SetSize(15, 12)
			end
			f.Text:SetText(text)
			f.Text:SetPoint(Points[direction+pad], f, 0, 0)
			pad = pad + 1
		end
		self.EndPoint1.Diametric = self.EndPoint2
		self.EndPoint2.Diametric = self.EndPoint1
		return f
	end

	local function GridCreateAxial(direction, x, y, id)
		local f = GridParent
		local table = f.Lines[direction]
		local count = #table
		local line
		if id <= count then
			line = table[id]
			GridAnchorLine(line, direction, x, y)
			return line
		end
		line = GridParent:CreateLine(nil, "BACKGROUND", nil, 0)
		tinsert(table, line)
		line.Id = id
		line.Direction = direction
		line:SetThickness(1)
		count = floor(#table/2)+1
		GridCreateEndPoints(line, direction, count)
		GridAnchorLine(line, direction, x, y)
		return line
	end

	local function GridResetLines(direction, count, xstep, ystep)
		for i=2, count do
			GridCreateAxial(direction, xstep, ystep, i)
		end
	end

	local function GridUpdate(self)
		GridSpacing = OptionsParent.Spacing:GetNumber()
		local xCount, yCount, xSpacing, ySpacing
		local Width = GetScreenWidth()
		local Height = GetScreenHeight()
		if GridByPixels then
			if GridSpacing < MinSpacing then
				GridSpacing = MinSpacing
				OptionsParent.Spacing:SetText(MinSpacing)
			end
			if GridSpacing > MaxSpacing then
				GridSpacing = MaxSpacing
				OptionsParent.Spacing:SetText(MaxSpacing)
			end
			xSpacing, ySpacing = GridSpacing, GridSpacing
			xCount = floor(Height/xSpacing)
			yCount = floor(Width/ySpacing)
			if xCount % 2 == 0 then
				xCount = xCount + 1
			end
			if yCount % 2 == 0 then
				yCount = yCount + 1
			end
		else
			if GridSpacing % 2 ~= 0 then
				GridSpacing = GridSpacing - 1
				OptionsParent.Spacing:SetText(GridSpacing)
			end
			if GridSpacing < MinLines  then
				GridSpacing = MinLines
				OptionsParent.Spacing:SetText(MinLines)
			end
			if GridSpacing > MaxLines then
				GridSpacing = MaxLines
				OptionsParent.Spacing:SetText(MaxLines)
			end
			xCount, yCount = GridSpacing, GridSpacing
			xSpacing = Height / xCount
			ySpacing = Width / yCount
			xCount = xCount - 1
			yCount = yCount - 1
		end
		if not GridParent then
--			GridParent = CreateFrame('Frame', nil, OptionsParent)
			GridParent = CreateFrame('Frame', nil, gridparent)
			GridParent:SetAllPoints(UIParent)
			GridParent:SetFrameStrata("TOOLTIP")
			GridParent:SetFrameLevel(1)
			GridParent.Lines = { [Horizontal]={},  [Vertical]={} }
			GridParent.Count = {}
			local line = GridCreateAxial(Horizontal, 0, 0, 1)
			line = GridCreateAxial(Vertical, 0, 0, 1)
		end
		GridParent.Count[Horizontal] = xCount
		GridParent.Count[Vertical] = yCount
		GridProcessLines(All, GridSetVisible, false)
		if self.Id == All or self.Id == Horizontal then
			GridResetLines(Horizontal, xCount, 0, xSpacing)
		end
		if self.Id == All or self.Id == Vertical then
			GridResetLines(Vertical, yCount, ySpacing, 0)
		end
		OptionsParent.ClearAll:SetEnabled(true)
		GridParent:Show()
	end

	local function GridOptionOnClick(self, preserve)
		local Id = self:GetParent().Id
		OptionsParent.Reset:SetEnabled(false)
		OptionsParent.DisplayTicks:SetEnabled(true)
		if not preserve and Id and Id == self.Id then
			self:GetParent().Id = nil
			PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_OFF)
			GridParent:Hide()
			if GridParent then
				OptionsParent.ClearAll:SetEnabled(false)
			end
			return
		end
		self:GetParent().Id = self.Id
		OptionsParent.Spacing:ClearFocus()
		PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON)
		GridUpdate(self)
	end

--------------------------------------------------------------------------------------------------------------------------------
--							     Control			 				      --
--------------------------------------------------------------------------------------------------------------------------------

	f.Spacing = CreateFrame("EditBox", "$parent:GridCount", f, "InputBoxTemplate")
	f.Spacing:SetSize(40, 20)
	f.Spacing:SetPoint(unpack(point))
	AddTooltips(f.Spacing, gTooltips)
	f.Spacing:SetMaxLetters(3)
	f.Spacing:SetAutoFocus(false)
	f.Spacing:ClearFocus()
	f.Spacing.Id = All
	f.Spacing:SetText(gSettings.initspacing)
	f.Spacing:SetScript("OnEnterPressed", function(self)
		GridSpacing = self:GetNumber()
		GridOptionOnClick(self, true)
	end)
	f.Spacing.Label = f:CreateFontString()
	f.Spacing.Label:SetFontObject(GameFontNormalSmall)
--	f.Spacing.Label:SetTextColor(1, 1, 1)
	f.Spacing.Label:SetText("| |     ><")
	f.Spacing.Label:SetPoint("BOTTOMLEFT", f.Spacing, "TOPLEFT", -3, 0)


	GridByPixels = gSettings.usepixels
	f.InPixels = CreateFrame("Button", "$parent:InPixels", f, "UIPanelButtonTemplate")
	f.InPixels:SetSize(24, 22)
	f.InPixels:SetText("| |")
	AddTooltips(f.InPixels, gTooltips)
	f.InPixels:SetPoint("LEFT", f.Spacing, "RIGHT", 4, 0)
	f.InPixels:SetScript("OnClick", function(self)
		self:GetParent().Spacing:ClearFocus()
		GridByPixels = not GridByPixels
		if GridByPixels then
			self:SetText("><")
		else
			self:SetText("| |")
		end
		if GridParent and GridParent:IsShown() then
			GridOptionOnClick(self:GetParent(), true)
		end
	end)
	f.DisplayTicks = CreateFrame("Button", "$parent:DisplayTicks", f, "UIPanelButtonTemplate")
	f.DisplayTicks:SetSize(22, 22)
	f.DisplayTicks:SetPoint("LEFT", f.InPixels, "RIGHT", 4, 0)
	f.DisplayTicks:SetText(": :")
	AddTooltips(f.DisplayTicks, gTooltips)
	f.DisplayTicks:SetEnabled(false)
	f.DisplayTicks:SetScript("OnClick", function(self)
			DisplayTicks = not DisplayTicks
			GridProcessLines(All, GridSetVisible, true)
	end)
	
	f.ClearAll = CreateFrame("Button", "$parent:ClearAll", f, "UIPanelButtonTemplate")
	f.ClearAll:SetSize(22, 22)
	f.ClearAll:SetPoint("LEFT", f.DisplayTicks, "RIGHT", 4, 0)
	f.ClearAll:SetText("x")
	AddTooltips(f.ClearAll, gTooltips)
	f.ClearAll:SetEnabled(false)
	f.ClearAll:SetScript("OnClick", function(self)
		self:GetParent().Spacing:ClearFocus()
		self:GetParent().Id = nil
		self:SetEnabled(false)
		GridParent:Hide()
	end)
	
	f.ToggleAll = CreateFrame("Button", "$parent:ToggleAll", f, "UIPanelButtonTemplate")
	f.ToggleAll:SetSize(22, 22)
	f.ToggleAll:SetPoint("LEFT", f.ClearAll, "RIGHT", 10, 0)
	f.ToggleAll:SetText("+")
	AddTooltips(f.ToggleAll, gTooltips)
	f.ToggleAll.Id = All
	f.ToggleAll:SetScript("OnClick", function(self)
		GridOptionOnClick(self)
	end)
	if block then
		f.ToggleAll:SetPoint("TOPLEFT", f.Spacing, "BOTTOMLEFT", -8, -5)
	else
		f.ToggleAll:SetPoint("LEFT", f.ClearAll, "RIGHT", 10, 0)
	end
	
	f.ToggleVert = CreateFrame("Button", "$parent:ToggleVertical", f, "UIPanelButtonTemplate")
	f.ToggleVert:SetSize(22, 22)
	f.ToggleVert:SetPoint("LEFT", f.ToggleAll, "RIGHT", 4, 0)
	f.ToggleVert:SetText("^")
	AddTooltips(f.ToggleVert, gTooltips)
	f.ToggleVert.Id = Vertical
	f.ToggleVert:SetScript("OnClick", function(self)
		GridOptionOnClick(self)
	end)

	f.ToggleHoriz = CreateFrame("Button", "$parent:ToggleHorizontal", f, "UIPanelButtonTemplate")
	f.ToggleHoriz:SetSize(22, 22)
	f.ToggleHoriz:SetPoint("LEFT", f.ToggleVert, "RIGHT", 4, 0)
	f.ToggleHoriz:SetText(">")
	AddTooltips(f.ToggleHoriz, gTooltips)
	f.ToggleHoriz.Id = Horizontal
	f.ToggleHoriz:SetScript("OnClick", function(self)
		GridOptionOnClick(self)
	end)

	f.Reset = CreateFrame("Button", "$parent:Reset", f, "UIPanelButtonTemplate")
	f.Reset:SetSize(22, 22)
	f.Reset:SetPoint("LEFT", f.ToggleHoriz, "RIGHT", 4, 0)
	f.Reset:SetText("#")
	AddTooltips(f.Reset, gTooltips)
	f.Reset:SetEnabled(false)
	f.Reset:SetScript("OnClick", function(self)
		GridOptionOnClick(self:GetParent(), true)
	end)
	return f.Spacing, f.ClearAll, f.ToggleAll
end

--------------------------------------------------------------------------------------------------------------------------------
--							     Settings			 				      --
--------------------------------------------------------------------------------------------------------------------------------

function AzGridOptionsInitialise(f, gridframe, settings, tooltips, defaults, point)
	if not gridframe then 
		print("No Grid Control Parent Frame Supplied!")
		return
	end

	local gSettings = settings
	local gTooltips = tooltips
	local HighStrata = "TOOLTIP"
	local CPStrata = ColorPickerFrame:GetFrameStrata()
	local TTStrata = GameTooltip:GetFrameStrata()
	local TTLevel = GameTooltip:GetFrameLevel()
	
	local All, Horizontal, Vertical = 0, 1, 3

	local function DartColor(color)
		local Color = { r=color[1], g=color[2], b=color[3] }
		return Color
	end

	local function ReColor(color, r, g, b)
		color[1] = r
		color[2] = g
		color[3] = b
	end

	local function ColorPickerOk(self)
		ColorPickerFrame:SetFrameStrata(CPStrata)
	end
	
	local function ColorPickerCancelled(self)
		_G[ColorPickerFrame.colorBox]:SetBackdropColor(unpack(ColorPickerFrame.previousValues))
		ReColor(gSettings.colors[ColorPickerFrame.color], unpack(ColorPickerFrame.previousValues))
		gridframe.GridProcessLines(All, gridframe.GridSetColor)
		ColorPickerFrame:SetFrameStrata(CPStrata)
	end
	
	local function ColorPickerColorChanged(self)
		local r, g, b = ColorPickerFrame:GetColorRGB()
		_G[ColorPickerFrame.colorBox]:SetBackdropColor(r, g, b)
		ReColor(gSettings.colors[ColorPickerFrame.color], r, g, b)
		gridframe.GridProcessLines(All, gridframe.GridSetColor)
	end

	local function ColorPickerOnClick(self)
		local color = DartColor(gSettings.colors[self.color])
		ColorPickerFrame.previousValues = { unpack(gSettings.colors[self.color]) }
--		ColorPickerFrame.cancelFunc = ColorPickerCancelled
--		ColorPickerFrame.opacityFunc = ColorPickerOk
--		ColorPickerFrame.func = ColorPickerColorChanged
--		ColorPickerFrame:SetColorRGB(color.r, color.g, color.b)
		ColorPickerFrame.colorBox = self
		ColorPickerFrame.color = self.color

		if ColorPickerFrame.SetupColorPickerAndShow then  -- 10.2.5 fix ColorSelect widget moved from top level to "Frame".Content.ColorPicker
			local info = UIDropDownMenu_CreateInfo()
			info.r, info.g, info.b = color.r, color.g, color.b
			info.swatchFunc = DART_ColorPicker_ColorChanged
			info.cancelFunc = DART_ColorPicker_ColorCancelled
			ColorPickerFrame:SetupColorPickerAndShow(info)
			if ColorPickerFrame.Content then
				ColorPickerFrame.Content.ColorPicker:SetColorRGB(color.r, color.g, color.b)
			else
				ColorPickerFrame.ColorPicker:SetColorRGB(color.r, color.g, color.b)
			end
		else
			ColorPickerFrame.cancelFunc = DART_ColorPicker_ColorCancelled
			ColorPickerFrame.opacityFunc = DART_ColorPicker_ColorChanged
			ColorPickerFrame.func = DART_ColorPicker_ColorChanged
			ColorPickerFrame:SetColorRGB(color.r, color.g, color.b)
		end



		ColorPickerFrame:ClearAllPoints()
		ColorPickerFrame:SetFrameStrata(HighStrata)
		if self:GetParent():GetRight() < UIParent:GetWidth() / 2 then
			ColorPickerFrame:SetPoint("TOPLEFT", self, "TOPRIGHT", 10, 0)
		else
			ColorPickerFrame:SetPoint("TOPRIGHT", self, "TOPLEFT", -10, 0)
		end
		ColorPickerFrame:Show()
	end

	local function CreateColorPicker(name, parent)
		local backdrop = {
			bgFile="Interface\\AddOns\\DiscordLibrary\\PlainBackdrop",
			edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
			tile=true,
			edgeSize=16,
			tileSize=16,
			insets = {
				left = 5,
				right = 5,
				top = 5,
				bottom = 5,
			},
		}
		local f = CreateFrame("Button", "$parent:"..name, parent, BackdropTemplateMixin and "BackdropTemplate")
		f:SetBackdrop(backdrop)
		f:SetSize(25, 25)
		f:SetScript("OnClick", ColorPickerOnClick)
		f.color = name
		DL_Init_ColorPicker(f, DartColor(gSettings.colors[name]))
		f.Label = f:CreateFontString()
		f.Label:SetFontObject(GameFontNormal)
		f.Label:SetTextColor(1, 1, 1)
		f.Label:SetPoint("CENTER")
		function f:SetLabel(text) self.Label:SetText(text) end
		return f
	end
	
	local Pickers = {}
	f.Centers = CreateColorPicker("Centers", f)
	tinsert(Pickers, f.Centers)
	f.Centers:SetPoint(unpack(point))
	f.Centers:SetLabel("+")
	AddTooltips(f.Centers, gTooltips)

	f.Lines = CreateColorPicker("Lines", f)
	tinsert(Pickers, f.Lines)
	f.Lines:SetPoint("LEFT", f.Centers, "RIGHT", 4, 0)
	f.Lines:SetLabel("|")
	AddTooltips(f.Lines, gTooltips)

	f.EndPoints = CreateColorPicker("EndPoints", f)
	tinsert(Pickers, f.EndPoints)
	f.EndPoints:SetPoint("LEFT", f.Lines, "RIGHT", 4, 0)
	f.EndPoints:SetLabel("++")
	AddTooltips(f.EndPoints, gTooltips)

	f.Texts = CreateColorPicker("Texts", f)
	tinsert(Pickers, f.Texts)
	f.Texts:SetPoint("LEFT", f.EndPoints, "RIGHT", 4, 0)
	f.Texts:SetLabel("12")
	AddTooltips(f.Texts, gTooltips)

	f.Highlight = CreateColorPicker("Highlight", f)
	tinsert(Pickers, f.Highlight)
	f.Highlight:SetPoint("LEFT", f.Texts, "RIGHT", 4, 0)
	f.Highlight:SetLabel("**")
	AddTooltips(f.Highlight, gTooltips)

	f.Defaults = CreateFrame("Button", "$parent:Defaults", f, "UIPanelButtonTemplate")
	f.Defaults:SetSize(25, 25)
	f.Defaults:SetPoint("LEFT", f.Highlight, "RIGHT", 15, 0)
	f.Defaults:SetText("@")
	AddTooltips(f.Defaults, gTooltips)
	f.Defaults:SetScript("OnClick", function(self)
		DL_Copy_Table(defaults, gSettings)
		gridframe.GridProcessLines(All, gridframe.GridSetColor)
		for k, v in pairs(Pickers) do
			DL_Init_ColorPicker(v, DartColor(gSettings.colors[v.color]))
		end
	end)

end
